import inspect
import queue
import importlib
from typing import Any, Type, Callable, Union, Optional, List, Dict, TypeVar, Tuple

from ..proxy import *
from ...utils import name_utils
from .base import BeanFactory
from .initializing_bean import InitializingBean
from ...injects.objects import Autowired
from ...utils.type_utils import is_basic_type

_T = TypeVar('_T', bound=Any)


class _Param:
    name: str
    cls: Any
    primary: bool

    def __init__(self, name: str = None, cls: Any = None, primary: bool = False):
        self.name = name
        self.cls = cls
        self.primary = primary

    def __lt__(self, other):
        return False


class SimpleBeanFactory(BeanFactory):
    """简单bean工厂"""

    def __init__(self):
        self._name_bean = {}
        self._type_bean = {}
        self._init_queue = queue.PriorityQueue()
        self._dependency_map = {}
        self._register_class_object_bean(name='simpleBeanFactory', obj=self)
        self._initialized = False

    def get_bean(self, name: str = None, required_type: Union[Type, Callable] = None) -> Any:
        if not name and not required_type:
            raise ValueError('bean名称和类型不能都为空')
        if name:
            return self._resolve_name_bean(name, required_type)
        return self._resolve_type_bean(required_type)

    def get_beans(self, required_type: Type[_T]) -> List[_T]:
        return [*(self._resolve_type_beans(required_type))]

    def _resolve_bean_from_beans(self, beans):
        if len(beans) >= 2:
            for bean in beans:
                # 有primary返回primary
                if bean.primary():
                    return bean
            if len(beans) >= 2:
                raise RuntimeError('获取的bean数量不能超过1个, 请增加名称/类型获取bean或给某个bean配置primary属性')
        if beans:
            return beans[0]
        return None

    def _resolve_name_bean(self, name: str, required_type: Union[Type, Callable] = None):
        beans = self._name_bean.get(name)
        if not beans:
            return None
        # 特定类型
        if required_type:
            # 优先取相同类型的
            bean = self._resolve_bean_from_beans([bean for bean in beans if bean.ioc_type() == required_type])
            if bean:
                return bean
            # 同类型的未找到再找父类
            bean = self._resolve_bean_from_beans(
                [bean for bean in beans if issubclass(bean.ioc_type(), required_type)])
            if bean:
                return bean
            return None

        return self._resolve_bean_from_beans(beans)

    def _resolve_type_bean(self, required_type=None):
        beans = self._resolve_type_beans(required_type=required_type)
        return self._resolve_bean_from_beans(beans)

    def _resolve_type_beans(self, required_type=None):
        beans = self._type_bean.get(required_type)
        if not beans:
            beans = []
            for _type in self._type_bean.keys():
                if issubclass(_type, required_type):
                    beans = [*beans, *self._type_bean[_type]]
        return beans

    def _register_function_bean(self, name: str, func, primary: bool = False) -> Any:
        name = name or name_utils.to_camel_case(func.__name__, upper_case=False)
        proxy = ClassBeanProxy(name=name, obj=self._new_function_bean(func), primary=primary)
        self._add_bean(proxy.ioc_name(), proxy.ioc_type(), proxy)
        return proxy

    def _register_class_object_bean(self, name, obj, primary: bool = False) -> Any:
        name = name or name_utils.to_camel_case(obj.__class__.__name__, upper_case=False)
        proxy = ClassBeanProxy(name=name, obj=obj, primary=primary)
        self._add_bean(proxy.ioc_name(), proxy.ioc_type(), proxy)
        return proxy

    def _register_class_bean(self, name: str, cls, primary: bool = False) -> Any:
        return self._register_class_object_bean(name, self._new_class_bean(cls), primary=primary)


    def _new_class_bean(self, cls) -> Any:
        params = self._extract_func_enable_autowired_args(cls)
        return cls(**params)

    def _new_function_bean(self, func) -> Any:
        params = self._extract_func_enable_autowired_args(func)
        return func(**params)

    def _extract_fun_depends_args(self, func) -> Tuple[dict, dict]:
        """抽取函数依赖参数信息."""
        signature = inspect.signature(func)
        params, name_params = {}, {}
        for name, parameter in signature.parameters.items():
            annotation = parameter.annotation
            default = parameter.default
            # Autowired参数依赖顺序处理
            is_bean_proxy_type = issubclass(type(default), BaseBeanProxy)
            # 默认值不为空 则 不注入参数
            if default is not parameter.empty and not is_bean_proxy_type:
                continue
            # 代理类型默认值且annotation为空则通过代理对象获取被代理类型
            if is_bean_proxy_type:
                if annotation is parameter.empty:
                    annotation = default.ioc_type()
                name_params[name] = default.ioc_name()
            # 没有类型 或 是基本类型 则 不注入参数
            if annotation is parameter.empty or is_basic_type(annotation):
                continue
            # 参数名称和类型映射
            # 非字符串类型则直接返回类型
            if not isinstance(annotation, str):
                params[name] = annotation
            else:
                # 动态加载类型
                params[name] = getattr(importlib.import_module(inspect.getmodule(func).__name__), annotation)

        return params, name_params

    def _extract_class_attr_depends_args(self, cls) -> Tuple[dict, dict]:
        """抽取类属性依赖参数信息."""
        members = [member for member in inspect.getmembers(cls) if issubclass(type(member[-1]), BaseBeanProxy)]
        return {member[0]: member[-1].ioc_type() for member in members}, {member[0]: member[-1].ioc_name()  for member in members}

    def _extract_func_enable_autowired_args(self, func) -> dict:
        """抽取函数可注入参数."""
        params, name_params = self._extract_fun_depends_args(func)
        if params:
            for name, cls in params.items():
                params[name] = Autowired(value=name_params.get(name), cls=cls)
        return params

    def register_bean(self, name: str, cls, primary: bool = False, order: int = 0, lazy=True) -> Any:
        # 注册bean懒加载, 在init方法执行bean创建逻辑
        if lazy and not self._initialized:
            self._init_queue.put((order, _Param(name, cls, primary)))
        else:
            self._register_bean(name, cls, primary)

    def _register_bean(self, name: str, cls, primary: bool = False) -> Any:
        if inspect.isfunction(cls):
            return self._register_function_bean(name, cls, primary=primary)
        if inspect.isclass(cls):
            return self._register_class_bean(name, cls, primary=primary)
        # 否则直接视为对象注册
        return self._register_class_object_bean(name, cls, primary=primary)

    def init(self):
        """初始化容器, 在bean都注册进工厂队列后执行, 同时解决多个容器间的依赖问题"""
        # 创建容器
        self._create_beans()
        # 初始化
        self._do_beans_init()
        self._initialized = True

    def _do_beans_init(self):
        # 对所有容器执行__post_construct__, InitializingBean.after_properties_set 初始化方法
        for beans in self._name_bean.values():
            for bean in beans:
                # 容器原始类型是类才能执行
                if inspect.isclass(bean.ioc_type()):
                    if hasattr(bean.ioc_type(), '__post_construct__'):
                        bean.__post_construct__()
                    if issubclass(bean.ioc_type(), InitializingBean):
                        bean.after_properties_set()

    def _create_beans(self):
        depends = []
        # 优先创建无依赖的bean
        while True:
            try:
                order, param = self._init_queue.get(block=False)
                # 没有依赖则可以直接注册bean
                depends_params, depends_name_params = self._get_depends_args(param)
                if not depends_params:
                    self._register_bean(param.name, param.cls, param.primary)
                    continue
                # 依赖入队列
                depends.append({'param': param, 'depends': depends_params, 'depends_name': depends_name_params})
            except queue.Empty:
                break

        if not depends:
            return
        # 创建依赖bean
        self._create_depends_beans(depends)

    def _create_depends_beans(self, depends: List[Dict[str, Any]]):
        # 处理非循环依赖
        while True:
            end = True
            next_depends = []
            for dep in depends:
                params = {}
                name_map = dep.get('depends_name') or {}
                for name, cls in dep['depends'].items():
                    # 获取依赖bean实例
                    bean_name = name_map.get(name)
                    if bean_name:
                        obj = self.get_bean(name=bean_name, required_type=cls)
                    else:
                        obj = self.get_bean(name=name, required_type=cls) or self.get_bean(required_type=cls)
                    # 依赖bean不存在则跳过处理下一个
                    if obj is None:
                        break
                    params[name] = obj
                # 获取的依赖数量与目标一致则可以创建bean
                if len(params) == len(dep['depends']):
                    self._register_bean(dep['param'].name, dep['param'].cls, dep['param'].primary)
                    end = False
                else:
                    next_depends.append(dep)
            depends = next_depends
            # 没有需要解决的普通依赖就退出
            if end:
                break

        if not depends:
            return

        # 处理循环依赖
        for dep in depends:
            self._register_bean(dep['param'].name, dep['param'].cls, dep['param'].primary)

    def _get_depends_args(self, param: _Param) -> Tuple[Optional[dict], Optional[dict]]:
        """判断要创建的bean是否存在依赖
        """
        name, cls, primary = param.name, param.cls, param.primary
        if inspect.isclass(cls):
            # 属性Autowired依赖
            args, name_args = self._extract_class_attr_depends_args(cls)
            # 构造函数依赖
            f_args, f_name_args = self._extract_fun_depends_args(cls)
            args.update(f_args)
            name_args.update(f_name_args)
            return args, name_args
        if inspect.isfunction(cls):
            return self._extract_fun_depends_args(cls)
        # 对象注入无需依赖
        return None, None

    def _add_bean(self, name: str, _type, bean: Any):
        self._add_name_bean(name, bean)
        self._add_type_bean(_type, bean)

    def _add_name_bean(self, name: str, bean: Any):
        beans = self._name_bean.get(name) or []
        beans.append(bean)
        self._name_bean[name] = beans

    def _add_type_bean(self, _type, bean: Any):
        beans = self._type_bean.get(_type) or []
        beans.append(bean)
        self._type_bean[_type] = beans



import re


def extract_content(s):
    # 正则表达式模式，捕获 ${...} 中的内容
    pattern = r'\$\{(.*?)\}'
    # 使用 re.findall() 函数找到所有匹配的捕获组内容
    return re.findall(pattern, s)

def test_ioc_config():
    # 测试字符串
    input_string = "${a}asdh${b}sd"

    # 调用函数并打印结果
    result = extract_content(input_string)
    print(result)




import abc


class XXXService(abc.ABC):

    def do_anything(self):
        raise NotImplementedError


"""
xxx/service_impl.py
"""


class XXXAService(XXXService):

    def do_anything(self):
        print("XXXA")


class XXXBService(XXXAService):

    def do_anything(self):
        print("XXXB")


"""
xxx/config.py
"""
from seatools.ioc import Bean, ApplicationContext


@Bean(primary=True)
def new_xxxa_service():  # 涉及依赖可写在方法参数里
    return XXXAService()


@Bean(name='xxxb')
def new_xxxb_service():
    return XXXBService()


"""
xxx/main.py
"""
from seatools.ioc import run, Autowired


def start():
    # 启动ioc依赖
    run(scan_package_names='tests.test_ioc2', config_dir='./config')

def test_ioc():
    start()
    xxxa = Autowired(
        cls=XXXService)  # 由于XXXService有XXXAService和XXXBService两个子类, 并且均定义了IOC容器实例, 当仅传递类型时, 会自动扫描获取primary=True的Bean返回
    xxxa.do_anything()  # 输出：XXXA
    xxxb = Autowired('xxxb', cls=XXXService)  # 通过名称+类型获取 XXXBService IOC 容器实例
    xxxb.do_anything()  # 输出：XXXB


def test_ioc_beans():
    start()
    ctx = Autowired(cls=ApplicationContext)
    beans = ctx.get_beans(XXXService)
    for bean in beans:
        bean.do_anything()


@Bean(order=1)
class A:

    def hello(self):
        print("hello A")


@Bean(order=0)
class B:

    def hello(self):
        print("hello B")

def test_ioc_order():
    start()
    a = Autowired(cls=A)
    b = Autowired(cls=B)
    a.hello()
    b.hello()




import abc
from typing import Union, Type, Callable, Any


class BeanFactory(abc.ABC):
    """Bean工厂抽象类"""

    @abc.abstractmethod
    def get_bean(self, name: str = None, required_type: Union[Type, Callable] = None):
        """获取bean方法."""
        raise NotImplementedError

    def get_beans(self, required_type: Type):
        """获取某个类型的bean列表方法."""
        raise NotImplementedError

    @abc.abstractmethod
    def register_bean(self, name: str, cls, primary: bool = False, order: int = 0, **kwargs) -> Any:
        """bean注册方法."""
        raise NotImplementedError

    @abc.abstractmethod
    def init(self, **kwargs):
        """创建bean的初始化方法."""
        raise NotImplementedError



import asyncio
import json

import pytest

from seatools.cache import Cache, AsyncCache, TTLCache
from seatools.cache.ext import SqliteCache
from loguru import logger
from seatools.models import BaseModel


cache = Cache(cache=TTLCache(maxsize=100, ttl=60), debug=True)
async_cache = AsyncCache(cache=TTLCache(maxsize=100, ttl=60), debug=True)


class D(BaseModel):
    a: int


@cache.cacheable(key='func1_${a}_${b}')
def func1(a: int, b: str):
    return f'{a}_{b}'


@cache.cache_evict(key='func1_${a}_${b}')
def clear_func1_cache(a: int, b: str):
    pass


@cache.cache_put(key='func1_${a}_${b}')
def put_func1_cache(a: int, b: str):
    return f'{a}_{b}'


@cache.cache_put(key='func_${s.1}')
def put_list_cache(s: list):
    return json.dumps(s)


@cache.cache_put(key='func_${s.s}')
def put_list_error_cache(s: list):
    return json.dumps(s)


class A:

    def __init__(self, a=1):
        self.a = a


@cache.cache_put('object_${a.a}')
def put_a(a: A):
    pass

@cache.cache_put('object_${a.b}')
def put_b(a: A):
    pass


def test_cache():
    v1 = put_func1_cache(1, 's')
    v1 = func1(1, 's', D(a=123))
    logger.info("v1: {}", v1)
    clear_func1_cache(1, 's')
    v2 = func1(1, 's')
    logger.info("v2: {}", v2)
    put_list_cache([1,2,3])
    cache.put('x', None)
    cache.get('x')
    cache.clear()
    with pytest.raises((IndexError, ValueError)):
        put_list_cache([1])
    with pytest.raises(ValueError):
        put_list_error_cache([1, 's'])
    put_a(A(a=2))
    with pytest.raises((AttributeError, ValueError)):
        put_b(A(a=2))




@async_cache.cacheable(key='func1_${a}_${b}')
async def async_func1(a: int, b: str):
    return f'{a}_{b}'


@cache.cacheable(key='func1_${a}_${b}')
async def async_func2(a: int, b: str):
    return f'{a}_{b}'


@async_cache.cache_evict(key='func1_${a}_${b}')
async def async_clear_func1_cache(a: int, b: str):
    pass


@async_cache.cache_put(key='func1_${a}_${b}')
async def async_put_func1_cache(a: int, b: str):
    return f'{a}_{b}'


@cache.cache_evict(key='func1_${a}_${b}')
async def async_clear_func2_cache(a: int, b: str):
    pass


@cache.cache_put(key='func1_${a}_${b}')
async def async_put_func2_cache(a: int, b: str):
    return f'{a}_{b}'


def test_async_cache():
    v1 = asyncio.run(async_put_func1_cache(1, 's'))
    v1 = asyncio.run(async_func1(1, 's'))
    logger.info("v1: {}", v1)
    asyncio.run(async_clear_func1_cache(1, 's'))
    v2 = asyncio.run(async_func1(1, 's'))
    logger.info("v2: {}", v2)


def test_async_cache2():
    v1 = asyncio.run(async_put_func2_cache(1, 's'))
    v1 = asyncio.run(async_func2(1, 's'))
    logger.info("v1: {}", v1)
    asyncio.run(async_clear_func2_cache(1, 's'))
    v2 = asyncio.run(async_func2(1, 's'))
    logger.info("v2: {}", v2)


def test_sqlite_cache(tmp_path):
    ori_sqlite_cache = SqliteCache(conn_url=f'sqlite:///tmp/cache.db', maxsize=100)
    sqlite_cache = Cache(cache=ori_sqlite_cache, debug=True)
    sqlite_cache.put('a', 1234)
    v = sqlite_cache.get('a')
    logger.info('a: {}', v)
    sqlite_cache.evict('a')
    v = sqlite_cache.get('a')
    logger.info('a: {}', v)
    with pytest.raises(KeyError):
        logger.info(ori_sqlite_cache['xxx'])



import importlib
from typing import List, Union
from threading import RLock

from loguru import logger

from seatools.ioc.context import ApplicationContext
from seatools.ioc.utils import reflect_utils
from seatools.ioc.base import new_bean_factory, get_bean_factory
from seatools.ioc.beans.factory import BeanFactory
from seatools.ioc.config import load_config


class _Properties:
    first = True
    lock = RLock()
    loaded_modules = set()


def run(scan_package_names: Union[List[str], str], config_dir: str = None, factory: BeanFactory = None,
        exclude_modules: List[str] = None):
    """启动pyspring, 一个进程中重复启动仅第一次生效

    Args:
        scan_package_names: 需要扫描的包, 支持多个包, 示例: xxx.xxx; xxx
        config_dir: 配置目录, 若不为空则加将config_dir目录下的 application.yml, application-{env}.yml 文件(如果存在) 加载到系统配置中
        同时支持@Value, @Configuration装饰器通过IOC注入是属性和bean
        factory: bean工厂, 可自己实现, 默认走SimpleBeanFactory
        exclude_modules: 需要过滤的模块列表
    """
    # 启动加锁
    with _Properties.lock:
        if not scan_package_names:
            raise ValueError('包名不能为空')
        if isinstance(scan_package_names, str):
            scan_package_names = [scan_package_names]
        # 初始化工厂
        if _Properties.first:
            bean_factory = new_bean_factory(factory=factory)
            # 有配置则初始化配置
            if config_dir:
                load_config(config_dir=config_dir)

            # 注册Environment
            from seatools.ioc.environment import Environment
            bean_factory.register_bean(name='environment', cls=Environment, primary=True, lazy=False)

            # 注册 ApplicationContext
            bean_factory.register_bean(name='applicationContext', cls=ApplicationContext(bean_factory), primary=True,
                                       lazy=False)

            # 添加starters包优先自动加载
            scan_package_names = ['seatools.ioc.starters'] + scan_package_names

            _Properties.first = False
        else:
            bean_factory = get_bean_factory()

        for scan_package_name in scan_package_names:
            # 加载包下所有某个和bean, 并注入到工厂
            modules = reflect_utils.get_all_py_modules(scan_package_name)
            for module in modules:
                if not exclude_modules:
                    _load_module(module)
                    continue
                for exclude_module in exclude_modules:
                    if module.startswith(exclude_module):
                        break
                else:
                    _load_module(module)
        # 初始化创建bean实例
        bean_factory.init()
        # 释放加载资源
        _release()


def _release():
    # 释放资源
    _Properties.loaded_modules.clear()


def _load_module(module):
    if module in _Properties.loaded_modules:
        return
    # 防止出现递归无限加载, 优先放入记录再加载
    _Properties.loaded_modules.add(module)
    try:
        importlib.import_module(module)
    except ModuleNotFoundError:
        logger.warning('Module {} not found.Cannot to import by seatools.ioc'.format(module))



from seatools.builders import HtmlBuilder, MarkdownBuilder


def test_html_builder():
    b = HtmlBuilder(auto_add_html_tag=True)
    b.h1('标题h1').h2('标题h2').h3('标题h3').h4('标题h4').h5('标题h5').h6('标题h6').br().b('标签b').p('啦啦啦').em(
        '标签em').i('标签i').blockquote('标签blockquote').code('标签code').pre('标签pre').img(
        'https://blog.dragons96.com/logo.png').strong('标签strong').custom_tag('div', '自定义标签').no_tag('hello').div(
        HtmlBuilder().p('11111')).a('测试链接', 'https://www.baidu.com')._common_tag()
    html = b.build()
    assert html, 'html为空！'


def test_markdown_builder():
    m = MarkdownBuilder()
    m.h1('标题h1').h2('标题h2').h3('标题h3').h4('标题h4').h5('标题h5').h6('标题h6').hr().b('标签b').p('啦啦啦').em(
        '标签em').i('标签i').blockquote('标签blockquote').code('标签code').pre('标签pre').img(
        'https://blog.dragons96.com/logo.png').strong('标签strong').no_tag(
        'hello').a('测试链接', 'https://www.baidu.com').del_('删除线')
    markdown = m.build()
    assert markdown, 'markdown为空！'
